@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2009 Jeffrey Stedfast
+ *  Copyright (C) 2000-2011 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -26,10 +26,11 @@
 #include <string.h>
 #include <ctype.h>
 
+#include "gmime-stream-mem.h"
 #include "gmime-common.h"
 #include "gmime-header.h"
+#include "gmime-events.h"
 #include "gmime-utils.h"
-#include "gmime-stream-mem.h"
 
 #include "list.h"
 
@@ -67,6 +68,7 @@ struct _GMimeHeader {
 struct _GMimeHeaderList {
 	GMimeStream *stream;
 	GHashTable *writers;
+	GMimeEvent *changed;
 	GHashTable *hash;
 	guint32 version;
 	List list;
@@ -534,6 +536,7 @@ g_mime_header_list_new (void)
 	headers->hash = g_hash_table_new (g_mime_strcase_hash,
 					  g_mime_strcase_equal);
 	list_init (&headers->list);
+	headers->changed = g_mime_event_new (headers);
 	headers->stream = NULL;
 	headers->version = 0;
 	
@@ -568,10 +571,62 @@ g_mime_header_list_destroy (GMimeHeaderList *headers)
 	if (headers->stream)
 		g_object_unref (headers->stream);
 	
+	g_mime_event_destroy (headers->changed);
+	
 	g_slice_free (GMimeHeaderList, headers);
 }
 
 
+/**
+ * g_mime_header_list_clear:
+ * @headers: a #GMimeHeaderList
+ *
+ * Removes all of the headers from the #GMimeHeaderList.
+ **/
+void
+g_mime_header_list_clear (GMimeHeaderList *headers)
+{
+	GMimeHeader *header, *next;
+	
+	g_return_if_fail (headers != NULL);
+	
+	header = (GMimeHeader *) headers->list.head;
+	while (header->next) {
+		next = header->next;
+		g_mime_header_free (header);
+		header = next;
+	}
+	
+	g_hash_table_remove_all (headers->hash);
+	
+	g_mime_header_list_set_stream (headers, NULL);
+}
+
+
+/**
+ * g_mime_header_list_contains:
+ * @headers: a #GMimeHeaderList
+ * @name: header name
+ *
+ * Checks whether or not a header exists.
+ *
+ * Returns: %TRUE if the specified header exists or %FALSE otherwise.
+ **/
+gboolean
+g_mime_header_list_contains (const GMimeHeaderList *headers, const char *name)
+{
+	const GMimeHeader *header;
+	
+	g_return_val_if_fail (headers != NULL, FALSE);
+	g_return_val_if_fail (name != NULL, FALSE);
+	
+	if (!(header = g_hash_table_lookup (headers->hash, name)))
+		return FALSE;
+	
+	return TRUE;
+}
+
+
 /**
  * g_mime_header_list_prepend:
  * @headers: a #GMimeHeaderList
@@ -594,10 +649,7 @@ g_mime_header_list_prepend (GMimeHeaderList *headers, const char *name, const ch
 	list_append (&headers->list, (ListNode *) header);
 	g_hash_table_replace (headers->hash, header->name, header);
 	
-	if (headers->stream) {
-		g_object_unref (headers->stream);
-		headers->stream = NULL;
-	}
+	g_mime_header_list_set_stream (headers, NULL);
 }
 
 
@@ -625,10 +677,7 @@ g_mime_header_list_append (GMimeHeaderList *headers, const char *name, const cha
 	if (!g_hash_table_lookup (headers->hash, name))
 		g_hash_table_insert (headers->hash, header->name, header);
 	
-	if (headers->stream) {
-		g_object_unref (headers->stream);
-		headers->stream = NULL;
-	}
+	g_mime_header_list_set_stream (headers, NULL);
 }
 
 
@@ -702,10 +751,7 @@ g_mime_header_list_set (GMimeHeaderList *headers, const char *name, const char *
 		g_hash_table_insert (headers->hash, header->name, header);
 	}
 	
-	if (headers->stream) {
-		g_object_unref (headers->stream);
-		headers->stream = NULL;
-	}
+	g_mime_header_list_set_stream (headers, NULL);
 }
 
 
@@ -749,10 +795,7 @@ g_mime_header_list_remove (GMimeHeaderList *headers, const char *name)
 	list_unlink ((ListNode *) header);
 	g_mime_header_free (header);
 	
-	if (headers->stream) {
-		g_object_unref (headers->stream);
-		headers->stream = NULL;
-	}
+	g_mime_header_list_set_stream (headers, NULL);
 	
 	return TRUE;
 }
@@ -947,8 +990,6 @@ g_mime_header_list_register_writer (GMimeHeaderList *headers, const char *name,
  * @stream: a #GMimeStream
  *
  * Set the raw header stream.
- *
- * Since: 2.4.7
  **/
 void
 g_mime_header_list_set_stream (GMimeHeaderList *headers, GMimeStream *stream)
@@ -956,6 +997,9 @@ g_mime_header_list_set_stream (GMimeHeaderList *headers, GMimeStream *stream)
 	g_return_if_fail (stream == NULL || GMIME_IS_STREAM (stream));
 	g_return_if_fail (headers != NULL);
 	
+	if (headers->stream == stream)
+		return;
+	
 	if (stream)
 		g_object_ref (stream);
 	
@@ -963,6 +1007,8 @@ g_mime_header_list_set_stream (GMimeHeaderList *headers, GMimeStream *stream)
 		g_object_unref (headers->stream);
 	
 	headers->stream = stream;
+	
+	g_mime_event_emit (headers->changed, NULL);
 }
 
 
@@ -973,8 +1019,6 @@ g_mime_header_list_set_stream (GMimeHeaderList *headers, GMimeStream *stream)
  * Gets the raw stream representing @headers.
  *
  * Returns: a #GMimeStream if set or %NULL otherwise.
- *
- * Since: 2.4.7
  **/
 GMimeStream *
 g_mime_header_list_get_stream (GMimeHeaderList *headers)
@@ -985,45 +1029,10 @@ g_mime_header_list_get_stream (GMimeHeaderList *headers)
 }
 
 
-/**
- * g_mime_header_list_set_raw:
- * @headers: a #GMimeHeaderList
- * @raw: raw mime part header
- *
- * Set the raw header.
- *
- * Deprecated:2.4.7: Use g_mime_header_list_set_stream() instead.
- **/
-void
-g_mime_header_list_set_raw (GMimeHeaderList *headers, const char *raw)
+GMimeEvent *
+_g_mime_header_list_get_changed_event (GMimeHeaderList *headers)
 {
-	g_return_if_fail (headers != NULL);
+	g_return_val_if_fail (headers != NULL, NULL);
 	
-	if (headers->stream)
-		g_object_unref (headers->stream);
-	
-	if (raw)
-		headers->stream = g_mime_stream_mem_new_with_buffer (raw, strlen (raw));
-	else
-		headers->stream = NULL;
-}
-
-
-/**
- * g_mime_header_list_has_raw:
- * @headers: a #GMimeHeaderList
- *
- * Gets whether or not a raw header has been set on @headers.
- *
- * Returns: %TRUE if a raw header is set or %FALSE otherwise.
- *
- * Deprecated:2.4.7: Use g_mime_header_list_get_stream() and check
- * against NULL instead.
- **/
-gboolean
-g_mime_header_list_has_raw (const GMimeHeaderList *headers)
-{
-	g_return_val_if_fail (headers != NULL, FALSE);
-	
-	return headers->stream ? TRUE : FALSE;
+	return headers->changed;
 }
